@@ -1,5 +1,9 @@
 # -*- coding: utf-8 -*-
 """Top-level package for twinthread."""
+import io
+
+from _plotly_utils.utils import PlotlyJSONEncoder
+
 from .jupyter import register_jupyter
 from .string import task_string_to_context
 
@@ -9,9 +13,20 @@ __version__ = "0.4.0"
 
 import re
 import json
+import simplejson
 import requests
 import pandas as pd
 from io import StringIO
+from matplotlib.figure import Figure
+import matplotlib.pyplot as plt
+
+
+def subplot_to_fig(subplot):
+    if isinstance(subplot, Figure):
+        return subplot
+    fig = plt.figure()
+    fig.axes.append(subplot)
+    return fig
 
 
 def do_login(base_url, username):
@@ -59,6 +74,31 @@ def mkdir(name):
         os.mkdir(name)
 
 
+def plotlyfig2json(fig, fpath=None, context_parameters={}):
+    """
+    Serialize a plotly figure object to JSON so it can be persisted to disk.
+    Figures persisted as JSON can be rebuilt using the plotly JSON chart API:
+
+    http://help.plot.ly/json-chart-schema/
+
+    If `fpath` is provided, JSON is written to file.
+
+    Modified from https://github.com/nteract/nteract/issues/1229
+    """
+    redata = json.loads(json.dumps(fig["data"], cls=PlotlyJSONEncoder))
+    relayout = json.loads(json.dumps(fig["layout"], cls=PlotlyJSONEncoder))
+    context_parameters = simplejson.dumps(context_parameters, ignore_nan=True)
+    fig_json = json.dumps(
+        {"data": redata, "layout": relayout, "contextParameters": context_parameters}
+    )
+
+    if fpath:
+        with open(fpath, "w") as outfile:
+            json.dump(fig_json, outfile)
+    else:
+        return fig_json
+
+
 def filter_results(results, search_text="", keys=default_keys):
     return [
         select_keys(result, keys)
@@ -90,7 +130,6 @@ class TwinThreadClient:
 
         headers = {"Authorization": f"Bearer {self.__access_token}"}
         data = {**self.__context, **body}
-        print("body", data)
         response = requests.post(
             f"{self.__base_url}/api{route}", data=data, headers=headers
         )
@@ -258,17 +297,27 @@ class TwinThreadClient:
 
     def __save_remote_file(self, type, name, contents):
         blob_name = re.sub(r"\W+", "", name).lower()
+        file_suffix = ".json" if type == "json" else ".png" if type == "png" else ""
         organization_name = "demo"  # this only works with masterdb right now.
-        blob_name = f"{self.__context['assetModelId']}_{self.__context['taskId']}_{blob_name}_org{organization_name}_i0"
+        blob_name = f"{self.__context['assetModelId']}_{self.__context['taskId']}_{blob_name}_org_i0{file_suffix}"
 
-        response_type = "text/csv" if type is "table" else "arraybuffer"
+        response_type = (
+            "text/csv"
+            if type is "table"
+            else "application/json"
+            if type is "json"
+            else "arraybuffer"
+        )
         content_type = "b" if response_type is "arraybuffer" else ""
 
         self.__auth_check()
 
         filename = type + "_" + name
         file = open(filename, "w" + content_type)
-        file.write(contents)
+        try:
+            file.write(contents.read())
+        except Exception as e:
+            file.write(contents)
         file.close()
         from requests_toolbelt import MultipartEncoder
 
@@ -289,6 +338,30 @@ class TwinThreadClient:
         p = requests.post(
             f"{self.__base_url}/api/Model/ImportTaskData", data=m, headers=headers
         )
+        print(p.status_code)
+
+    def save_plotly(self, fig, name=""):
+        json = plotlyfig2json(fig)
+
+        if self.__request_id is not None:
+            self.__save_local_file("json", name, json)
+        else:
+            self.__save_remote_file("json", name, json)
+
+        print("Saved plotly", name)
+
+    def save_image(self, fig, name=""):
+        fig = subplot_to_fig(fig)
+        buf = io.BytesIO()
+        fig.savefig(buf, format="png")
+        buf.seek(0)
+
+        if self.__request_id is not None:
+            self.__save_local_file("png", name, buf)
+        else:
+            self.__save_remote_file("png", name, buf)
+
+        print("Saved image", name)
 
     def save_table(self, table, name=""):
         if not isinstance(table, pd.DataFrame):
@@ -304,5 +377,8 @@ class TwinThreadClient:
 
         print("Saved table", name)
 
+    def save_output_dataset(self, table):
+        self.save_table(table, "dataset")
+
 
 register_jupyter()
